/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2017 Antoine Martin <antoine@nagafix.co.uk>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-timestamp
 *
 * Expose timestamps using an absolute monotonic clock.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! timestamp ! fakesink
 * ]|
 * </refsect2>
 */

#define VERSION "0.1.0"

#include <gst/gst.h>

#include "gsttimestamp.h"

GST_DEBUG_CATEGORY_STATIC (gst_timestamp_debug);
#define GST_CAT_DEFAULT gst_timestamp_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_DELTA
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

#define gst_timestamp_class parent_class
G_DEFINE_TYPE (GstTimestampFilter, gst_timestamp, GST_TYPE_ELEMENT);

static void gst_timestamp_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_timestamp_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_timestamp_sink_event (GstPad * pad, GstObject * parent, GstEvent * event);
static GstFlowReturn gst_timestamp_chain (GstPad * pad, GstObject * parent, GstBuffer * buf);


/* initialize the timestamp class */
static void
gst_timestamp_class_init (GstTimestampFilterClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_timestamp_set_property;
  gobject_class->get_property = gst_timestamp_get_property;

  g_object_class_install_property (gobject_class, PROP_DELTA,
      g_param_spec_int64 ("delta", "Delta", "Delta between the buffer timestamps and the system monotonic clock",
          G_MININT64, G_MAXINT64, 0, G_PARAM_READABLE));

  gst_element_class_set_details_simple(gstelement_class,
    "Timestamp",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    " <antoine@nagafix.co.uk>");

  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_timestamp_init (GstTimestampFilter * filter)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_event_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_timestamp_sink_event));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_timestamp_chain));
  GST_PAD_SET_PROXY_CAPS (filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  GST_PAD_SET_PROXY_CAPS (filter->srcpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);

  filter->delta = 0;
}

static void
gst_timestamp_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
}

static void
gst_timestamp_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstTimestampFilter *filter = GST_TIMESTAMPFILTER (object);

  switch (prop_id) {
    case PROP_DELTA:
      g_value_set_int64 (value, filter->delta);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static gboolean
gst_timestamp_sink_event (GstPad * pad, GstObject * parent, GstEvent * event)
{
  return gst_pad_event_default (pad, parent, event);
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_timestamp_chain (GstPad * pad, GstObject * parent, GstBuffer * buf)
{
  GstTimestampFilter *filter;

  filter = GST_TIMESTAMPFILTER (parent);
  GstClockTime pts = GST_BUFFER_PTS(buf);

  struct timespec ts;
  clock_gettime(CLOCK_MONOTONIC, &ts);
  ulong mts = ts.tv_sec*1000000000 + ts.tv_nsec;
  //ulong delta = mts-pts;
  //g_print ("pts=%li, mts=%li, delta=%li\n", pts, mts, delta);
  filter->delta = mts-pts;
  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, buf);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
timestamp_init (GstPlugin *timestamp)
{
  /* debug category for fltering log messages
   */
  GST_DEBUG_CATEGORY_INIT (gst_timestamp_debug, "timestamp",
      0, "timestamp interceptor");

  return gst_element_register (timestamp, "timestamp", GST_RANK_NONE,
      GST_TYPE_TIMESTAMPFILTER);
}

#define PACKAGE "timestamp"

/* gstreamer looks for this structure to register timestamp
 *
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    timestamp,
    "timestamp",
    timestamp_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
